// ----------------------------------------------------
// Made by: Peepss
// Warning! Do not use this code without my permission.
// ----------------------------------------------------

import settings from "../Settings.js";

import {
    world,
    system,
    ItemStack,
    EnchantmentTypes
} from "@minecraft/server";
import { Enchantments } from "classes/Enchantments.js";

class BackpackEventsManager {
    constructor() {
        this.listeners = {
            onHoldBackpack: [],
            onReleaseBackpack: [],
            onBackpackInventoryOpen: [],
            onBackpackInventoryClose: []
        };
        this.run();
    }

    //Use this to execute a code when a specific event is emitted.
    event(event, callback) {
        if (!this.listeners[event]) {
            throw new Error(`Evento desconhecido: ${event}`);
        }
        if (typeof callback === "function") {
            this.listeners[event].push(callback);
        } else {
            throw new Error("O callback deve ser uma função.");
        }
    }

    //Use this to emit a new event.
    emit(event, data) {
        if (!this.listeners[event]) {
            throw new Error(`Evento desconhecido: ${event}`);
        }
        for (const listener of this.listeners[event]) {
            listener(data);
        }
    }

    getAllDynamicProperties(target) {
        const propertyIdentifiers = target.getDynamicPropertyIds();

        if (propertyIdentifiers.length > 0) {
            const dynamicProperties = [];

            propertyIdentifiers.forEach(identifier => {
                const propertyData = target.getDynamicProperty(identifier);

                if (propertyData) {
                    dynamicProperties.push({
                        identifier: identifier,
                        propertyData: propertyData,
                    });
                }
            });

            return dynamicProperties;
        }

        return [];
    }

    //Set the property "peepss:item_data" if the backpack doesn't have any.
    itemData(player) {
        const selectedSlot = player.selectedSlotIndex;
        const playerInventory = player.getComponent("inventory").container;
        const playerSelectedItem = playerInventory.getItem(selectedSlot);

        if (!playerSelectedItem) return;

        const selectedItemData = playerSelectedItem.getDynamicProperty("peepss:item_data");

        if (selectedItemData) return;

        if (settings.backpacksTypeIds.includes(playerSelectedItem.typeId)) {
            let backpackId = Math.floor(Math.random() * 9000) + 1000;

            let itemDataInterface = {
                owner: player.name, //This is the player that owns the backpack.
                backpackId: backpackId, //This ID is used to attach the container entities to the backpack item that have spawned it.
                typeId: playerSelectedItem.typeId,
                slotsAmount: 27, //Amount of slots on backpack, minimum of 27 slots and maximum of 45.
                isBeingHeldByPlayer: false,
                pockets: {
                    //Normal pockets, gives the backpack +18 slots.
                    leatherPockets: false,
                    //Special pockets, gives the backpack +18 slots and a special hability.
                    goldenPockets: false,
                    endlessPockets: false
                }
            }

            playerSelectedItem.setDynamicProperty("peepss:item_data", JSON.stringify(itemDataInterface));
            playerInventory.setItem(selectedSlot, playerSelectedItem);

            let playerDataInterface = {
                lastItemHeld: {
                    slot: selectedSlot,
                    typeId: playerSelectedItem.typeId,
                    backpackId: backpackId,
                }
            }

            player.setDynamicProperty("peepss:player_data", JSON.stringify(playerDataInterface));

            this.emit("onHoldBackpack", { player: player, item: playerSelectedItem });
        }
    }

    //If the player releases a backpack, emit a event.
    onReleaseBackpack(player) {
        const selectedSlot = player.selectedSlotIndex;
        const playerInventory = player.getComponent("inventory").container;
        const playerSelectedItem = playerInventory.getItem(selectedSlot);

        const playerData = JSON.parse(player.getDynamicProperty("peepss:player_data") || "{}");

        const lastItemHeld = playerData.lastItemHeld || null;

        if (!playerSelectedItem) {
            this.handleItemRelease(player, lastItemHeld);
            playerData.lastItemHeld = null;
            player.setDynamicProperty("peepss:player_data", JSON.stringify(playerData));
            return;
        }

        const selectedItemDataRaw = playerSelectedItem.getDynamicProperty("peepss:item_data");
        if (!selectedItemDataRaw) {
            this.handleItemRelease(player, lastItemHeld);
            playerData.lastItemHeld = null;
            player.setDynamicProperty("peepss:player_data", JSON.stringify(playerData));
            return;
        }

        const selectedItemData = JSON.parse(selectedItemDataRaw);
        if (
            lastItemHeld &&
            lastItemHeld.backpackId === selectedItemData.backpackId &&
            lastItemHeld.typeId === playerSelectedItem.typeId &&
            lastItemHeld.slot === selectedSlot
        ) {
            return;
        }

        this.handleItemRelease(player, lastItemHeld);

        playerData.lastItemHeld = {
            slot: selectedSlot,
            typeId: playerSelectedItem.typeId,
            backpackId: selectedItemData.backpackId,
        };
        player.setDynamicProperty("peepss:player_data", JSON.stringify(playerData));
    }

    handleItemRelease(player, lastItemHeld) {
        if (lastItemHeld) {
            const playerInventory = player.getComponent("inventory").container;
            const item = playerInventory.getItem(lastItemHeld.slot);

            if (item) {
                const itemDataRaw = item.getDynamicProperty("peepss:item_data");
                if (itemDataRaw) {
                    const itemData = JSON.parse(itemDataRaw);

                    itemData.isBeingHeldByPlayer = false;
                    item.setDynamicProperty("peepss:item_data", JSON.stringify(itemData));
                    playerInventory.setItem(lastItemHeld.slot, item);
                    this.emit("onReleaseBackpack", { player, item: item });
                }
            }
        }
    }

    //If the player holds a backpack, emit a event.
    onHoldBackpack(player) {
        const selectedSlot = player.selectedSlotIndex;
        const playerInventory = player.getComponent("inventory").container;
        const playerSelectedItem = playerInventory.getItem(selectedSlot);

        const playerData = JSON.parse(player.getDynamicProperty("peepss:player_data") || "{}");

        if (!playerSelectedItem) return;

        const selectedItemDataRaw = playerSelectedItem.getDynamicProperty("peepss:item_data");
        if (!selectedItemDataRaw) {
            return;
        }

        const selectedItemData = JSON.parse(selectedItemDataRaw);

        if (selectedItemData.isBeingHeldByPlayer) {
            return;
        }

        selectedItemData.isBeingHeldByPlayer = true;
        playerSelectedItem.setDynamicProperty("peepss:item_data", JSON.stringify(selectedItemData));
        playerInventory.setItem(selectedSlot, playerSelectedItem);

        playerData.lastItemHeld = {
            slot: selectedSlot,
            typeId: playerSelectedItem.typeId,
            backpackId: selectedItemData.backpackId,
        };
        player.setDynamicProperty("peepss:player_data", JSON.stringify(playerData));

        this.emit("onHoldBackpack", { player, item: playerSelectedItem });
    }

    saveInventory(player, backpack, item) {
        const backpackInventory = backpack.getComponent("inventory").container;
        const itemData = JSON.parse(item.getDynamicProperty("peepss:item_data"));
        let isInventoryChanged = false;

        for (let slot = 0; slot < backpackInventory.size; slot++) {
            const currentItem = backpackInventory.getItem(slot);

            if ((itemData.pockets.endlessPockets && slot < 27) || (!itemData.pockets.endlessPockets)) {
                let itemDataInterface = {
                    typeId: "minecraft:air",
                    count: 1,
                    lore: [],
                    enchantments: [],
                    durability: null,
                    customName: null
                }
                if (currentItem) {
                    itemDataInterface.typeId = currentItem.typeId;
                    itemDataInterface.count = currentItem.amount;

                    const lore = currentItem.getLore();
                    if (lore && lore.length > 0) {
                        itemDataInterface.lore = lore;
                    }

                    const enchantments = Enchantments.getEnchants(currentItem);
                    if (enchantments.length > 0) {
                        itemDataInterface.enchantments = enchantments.map(enchant => ({
                            name: enchant.type.id,
                            level: enchant.level
                        }));
                    }

                    const durabilityComponent = currentItem.getComponent("durability");
                    if (durabilityComponent) {
                        itemDataInterface.durability = durabilityComponent.damage;
                    }

                    const nameTag = currentItem.nameTag;
                    if (nameTag) {
                        itemDataInterface.customName = nameTag;
                    }
                }

                const currentItemData = JSON.parse(item.getDynamicProperty(`peepss:inventory_slot_${slot}`) || '{}');
                if (JSON.stringify(itemDataInterface) !== JSON.stringify(currentItemData)) {
                    item.setDynamicProperty(`peepss:inventory_slot_${slot}`, JSON.stringify(itemDataInterface));
                    isInventoryChanged = true;
                }

                if (isInventoryChanged) {
                    const playerInventory = player.getComponent("inventory").container;

                    for (let pSlot = 0; pSlot < playerInventory.size; pSlot++) {
                        const playerCurrentSlot = playerInventory.getItem(pSlot);

                        if (playerCurrentSlot) {
                            const playerItemData = playerCurrentSlot.getDynamicProperty("peepss:item_data");
                            if (
                                playerItemData &&
                                JSON.parse(playerItemData).backpackId === JSON.parse(backpack.getDynamicProperty("peepss:backpack_data")).backpackId
                            ) {
                                playerInventory.setItem(pSlot, item);
                                break;
                            }
                        }
                    }
                }
            } else if (itemData.pockets.endlessPockets && slot > 26) {
                const endlessInventoryKey = `peepss:endless_inventory_${slot}`;
                let endlessItemData = {
                    typeId: "minecraft:air",
                    count: 1,
                    lore: [],
                    enchantments: [],
                    durability: null,
                    customName: null
                };

                if (currentItem) {
                    endlessItemData.typeId = currentItem.typeId;
                    endlessItemData.count = currentItem.amount;

                    const lore = currentItem.getLore();
                    if (lore && lore.length > 0) {
                        endlessItemData.lore = lore;
                    }

                    const enchantments = Enchantments.getEnchants(currentItem);
                    if (enchantments.length > 0) {
                        endlessItemData.enchantments = enchantments.map(enchant => ({
                            name: enchant.type.id,
                            level: enchant.level
                        }));
                    }

                    const durabilityComponent = currentItem.getComponent("durability");
                    if (durabilityComponent) {
                        endlessItemData.durability = durabilityComponent.damage;
                    }

                    const nameTag = currentItem.nameTag;
                    if (nameTag) {
                        endlessItemData.customName = nameTag;
                    }
                }

                const playerEndlessInventory = JSON.parse(player.getDynamicProperty(endlessInventoryKey) || '{}');
                if (JSON.stringify(endlessItemData) !== JSON.stringify(playerEndlessInventory)) {
                    player.setDynamicProperty(endlessInventoryKey, JSON.stringify(endlessItemData));
                }
            }
        }
    }

    loadInventory(player, backpack, item) {
        const itemProperties = this.getAllDynamicProperties(item);
        const backpackInventory = backpack.getComponent("minecraft:inventory").container;

        if (!backpackInventory) {
            console.error("Backpack inventory not found!");
            return;
        }

        console.log("Item properties:", itemProperties);

        // Carregando slots padrão da mochila
        itemProperties.forEach(property => {
            try {
                const { identifier, propertyData } = property;

                if (!identifier.includes("peepss:inventory_slot_")) return;

                const slot = parseInt(identifier.replace("peepss:inventory_slot_", ""));
                const parsedPropertyData = JSON.parse(propertyData);
                const { typeId, count, lore, enchantments, durability, customName } = parsedPropertyData;

                if (!typeId || typeId === "minecraft:air") {
                    backpackInventory.setItem(slot, null);
                    return;
                }

                const itemStack = new ItemStack(typeId, count);

                if (lore && lore.length > 0) {
                    if (typeof itemStack.setLore === "function") {
                        itemStack.setLore(lore);
                    }
                }

                if (enchantments) {
                    for (const enchant of enchantments) {
                        Enchantments.addEnchant(itemStack, { type: EnchantmentTypes.get(enchant.name), level: enchant.level });
                    }
                }

                if (durability !== null) {
                    const durabilityComponent = itemStack.getComponent("durability");

                    if (durabilityComponent) {
                        durabilityComponent.damage = durability;
                    }
                }

                if (customName) {
                    itemStack.nameTag = customName;
                }

                console.log(`Setting item in backpack slot ${slot}:`, itemStack);

                backpackInventory.setItem(slot, itemStack);
            } catch (error) {
                console.error(`Failed to load item in slot: ${property.identifier}`, error);
            }
        });

        // Carregando slots do inventário endlessPockets do jogador
        const itemData = JSON.parse(item.getDynamicProperty("peepss:item_data") || "{}");
        if (itemData.pockets && itemData.pockets.endlessPockets) {
            for (let slot = 27; slot < 45; slot++) {
                try {
                    const endlessInventoryKey = `peepss:endless_inventory_${slot}`;
                    const playerPropertyData = player.getDynamicProperty(endlessInventoryKey);

                    if (!playerPropertyData) {
                        console.log(`No data found for endlessPockets slot ${slot}`);
                        backpackInventory.setItem(slot, null);
                        continue;
                    }

                    const parsedPropertyData = JSON.parse(playerPropertyData);
                    const { typeId, count, lore, enchantments, durability, customName } = parsedPropertyData;

                    if (!typeId || typeId === "minecraft:air") {
                        backpackInventory.setItem(slot, null);
                        continue;
                    }

                    const itemStack = new ItemStack(typeId, count);

                    if (lore && lore.length > 0) {
                        if (typeof itemStack.setLore === "function") {
                            itemStack.setLore(lore);
                        }
                    }

                    if (enchantments) {
                        for (const enchant of enchantments) {
                            Enchantments.addEnchant(itemStack, { type: EnchantmentTypes.get(enchant.name), level: enchant.level });
                        }
                    }

                    if (durability !== null) {
                        const durabilityComponent = itemStack.getComponent("durability");

                        if (durabilityComponent) {
                            durabilityComponent.damage = durability;
                        }
                    }

                    if (customName) {
                        itemStack.nameTag = customName;
                    }

                    console.log(`Setting endlessPockets item in slot ${slot}:`, itemStack);

                    backpackInventory.setItem(slot, itemStack);
                } catch (error) {
                    console.error(`Failed to load endlessPockets slot ${slot}`, error);
                }
            }
        }
    }

    //Summons the backpack container entity and keeps teleporting it to the player.
    spawnBackpackEntity(player, item) {
        const { dimension, location } = player;

        var backpackEntity;
        const backpackItemData = JSON.parse(item.getDynamicProperty("peepss:item_data"));

        const { owner, backpackId, typeId, slotsAmount, isBeingHeldByPlayer, pockets } = backpackItemData;

        if (!pockets.leatherPockets && !pockets.goldenPockets && !pockets.endlessPockets) {
            backpackEntity = dimension.spawnEntity("peepss:backpack_container", location);
            backpackEntity.nameTag = "screen.backpack:no_pockets";
        } else if (pockets.leatherPockets) {
            backpackEntity = dimension.spawnEntity("peepss:pockets_backpack_container", location);
            backpackEntity.nameTag = "screen.backpack:leather_pockets";
        } else if (pockets.goldenPockets) {
            backpackEntity = dimension.spawnEntity("peepss:golden_pockets_backpack_container", location);
            backpackEntity.nameTag = "screen.backpack:golden_pockets";
        } else if (pockets.endlessPockets) {
            backpackEntity = dimension.spawnEntity("peepss:pockets_backpack_container", location);
            backpackEntity.nameTag = "screen.backpack:endless_pockets";
        }

        let backpackDataInterface = {
            owner: owner, //This is the player that owns the backpack.
            backpackId: backpackId, //This ID is used to attach the container entities to the backpack item that have spawned it.
            typeId: typeId,
            slotsAmount: slotsAmount, //Amount of slots on backpack, minimum of 27 slots and maximum of 45.
            isBeingHeldByPlayer: isBeingHeldByPlayer,
            pockets: {
                //Normal pockets, gives the backpack +18 slots.
                leatherPockets: pockets.leatherPockets,
                //Special pockets, gives the backpack +18 slots and a special hability.
                goldenPockets: pockets.goldenPockets,
                endlessPockets: pockets.endlessPockets
            }
        };

        backpackEntity.setDynamicProperty("peepss:backpack_data", JSON.stringify(backpackDataInterface));

        system.runInterval(() => {
            try {
                if (backpackEntity.isValid()) {
                    const selectedSlot = player.selectedSlotIndex;
                    const playerInventory = player.getComponent("inventory").container;
                    const playerSelectedItem = playerInventory.getItem(selectedSlot);

                    if (playerSelectedItem) {
                        const selectedItemDataRaw = playerSelectedItem.getDynamicProperty("peepss:item_data");

                        if (settings.pocketsTypeIds.includes(playerSelectedItem.typeId)) {
                            backpackEntity.triggerEvent("minecraft:despawn");
                            return;
                        }

                        if (selectedItemDataRaw) {
                            try {
                                const selectedItemData = JSON.parse(selectedItemDataRaw);

                                if (selectedItemData.backpackId !== backpackId) {
                                    backpackEntity.triggerEvent("minecraft:despawn");
                                    return;
                                }
                            } catch (error) {
                                console.error("Erro ao analisar os dados do item selecionado:", error);
                                backpackEntity.triggerEvent("minecraft:despawn");
                                return;
                            }
                        } else {
                            backpackEntity.triggerEvent("minecraft:despawn");
                            return;
                        }
                    } else {
                        backpackEntity.triggerEvent("minecraft:despawn");
                    }

                    this.saveInventory(player, backpackEntity, item);

                    if (backpackEntity.hasTag("inventoryOpen")) {
                        backpackEntity.removeTag("inventoryOpen");
                        this.emit("onBackpackInventoryOpen", { player, item });
                    }

                    if (backpackEntity.hasTag("inventoryClose")) {
                        backpackEntity.removeTag("inventoryClose");
                        this.emit("onBackpackInventoryClose", { player, item });
                    }

                    backpackEntity.teleport({
                        x: player.location.x,
                        y: player.location.y + 1.2,
                        z: player.location.z
                    }, { dimension: dimension });
                }
            } catch { }
        }, 0);

        return backpackEntity;
    }

    //Running all the events that need to be executed per tick
    run() {
        system.runInterval(() => {
            try {
                for (const player of world.getPlayers()) {
                    const selectedSlot = player.selectedSlotIndex;

                    const playerInventory = player.getComponent("inventory").container;
                    const playerSelectedItem = playerInventory.getItem(selectedSlot);

                    this.itemData(player);
                    this.onReleaseBackpack(player);
                    this.onHoldBackpack(player);

                    //If player dies with the backpack in hand, this will prevent the backpack from returning to their inventory with the lockMode "slot".
                    for (let slot = 0; slot < playerInventory.size; slot++) {
                        const currentItem = playerInventory.getItem(slot);

                        if (currentItem) {
                            if (currentItem.lockMode == "slot" && slot !== player.selectedSlotIndex) {
                                currentItem.lockMode = "none";
                                playerInventory.setItem(slot, currentItem);
                                break;
                            }
                        }
                    }
                }
            } catch { }
        }, 0)
    }
}

const backpackEventsManager = new BackpackEventsManager();

export default backpackEventsManager;